import sys
from dataclasses import dataclass
from datetime import datetime, time
from decimal import Decimal
from enum import Enum
from typing import Annotated, Any, Literal, Optional, TypedDict, Union
from uuid import UUID

import pytest
from serpyco_rs import Serializer, JsonSchemaBuilder
from serpyco_rs.metadata import Alias, CamelCase, Discriminator, Max, MaxLength, Min, MinLength, OmitNone


def test_to_json_schema():
    class EnumCls(Enum):
        a = 'a'

    @dataclass
    class InnerData:
        """Some important entity"""

        foo_filed: str

    @dataclass
    class OtherInnerData:
        """OtherInnerData"""

        optional_filed: Optional[InnerData] = None

    class AnotherInnerData(TypedDict):
        """AnotherInnerData"""

        bar: str

    @dataclass
    class Data:
        """Docs"""

        a: Annotated[int, Min(0), Max(10)]
        """A field with bounds"""
        b: float
        c: Decimal
        d: bool
        e: Annotated[str, MinLength(1), MaxLength(5)]
        f: UUID
        g: time
        h: datetime
        i: EnumCls
        j: InnerData
        k: list[int]
        l: tuple[int, str, InnerData]
        m: dict[str, int]
        n: Any
        o: Annotated[InnerData, CamelCase]
        p: Annotated[OtherInnerData, OmitNone]
        q: AnotherInnerData
        some_filed: Annotated[str, Alias('fooFiled')]

    serializer = Serializer(Data)

    assert serializer.get_json_schema() == {
        '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema.<locals>.Data',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                'tests.json_schema.test_convert.test_to_json_schema.<locals>.Data': {
                    'description': 'Docs',
                    'properties': {
                        'a': {
                            'description': 'A field with bounds',
                            'maximum': 10,
                            'minimum': 0,
                            'type': 'integer',
                        },
                        'b': {'type': 'number'},
                        'c': {
                            'oneOf': [{'type': 'string', 'format': 'decimal'}, {'type': 'number', 'format': 'decimal'}]
                        },
                        'd': {'type': 'boolean'},
                        'e': {'maxLength': 5, 'minLength': 1, 'type': 'string'},
                        'f': {'format': 'uuid', 'type': 'string'},
                        'g': {
                            'format': 'time',
                            'type': 'string',
                        },
                        'h': {
                            'format': 'date-time',
                            'type': 'string',
                        },
                        'i': {'enum': ['a'], 'type': 'string'},
                        'j': {
                            '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema.<locals>.InnerData'
                        },
                        'k': {'items': {'type': 'integer'}, 'type': 'array'},
                        'l': {
                            'maxItems': 3,
                            'minItems': 3,
                            'prefixItems': [
                                {'type': 'integer'},
                                {'type': 'string'},
                                {
                                    '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema.<locals>.InnerData'
                                },
                            ],
                            'type': 'array',
                        },
                        'm': {'additionalProperties': {'type': 'integer'}, 'type': 'object'},
                        'n': {},
                        'p': {
                            '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema.<locals>.OtherInnerData'
                        },
                        'o': {
                            '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema.<locals>.InnerData1'
                        },
                        'q': {
                            '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema.<locals>.AnotherInnerData'
                        },
                        'fooFiled': {'type': 'string'},
                    },
                    'required': [
                        'a',
                        'b',
                        'c',
                        'd',
                        'e',
                        'f',
                        'g',
                        'h',
                        'i',
                        'j',
                        'k',
                        'l',
                        'm',
                        'n',
                        'o',
                        'p',
                        'q',
                        'fooFiled',
                    ],
                    'type': 'object',
                },
                'tests.json_schema.test_convert.test_to_json_schema.<locals>.InnerData': {
                    'description': 'Some important entity',
                    'properties': {'foo_filed': {'type': 'string'}},
                    'required': ['foo_filed'],
                    'type': 'object',
                },
                'tests.json_schema.test_convert.test_to_json_schema.<locals>.InnerData2': {
                    'description': 'Some important entity',
                    'properties': {'foo_filed': {'type': 'string'}},
                    'required': ['foo_filed'],
                    'type': 'object',
                },
                'tests.json_schema.test_convert.test_to_json_schema.<locals>.InnerData1': {
                    'description': 'Some important entity',
                    'properties': {'fooFiled': {'type': 'string'}},
                    'required': ['fooFiled'],
                    'type': 'object',
                },
                'tests.json_schema.test_convert.test_to_json_schema.<locals>.OtherInnerData': {
                    'description': 'OtherInnerData',
                    'properties': {
                        'optional_filed': {
                            'anyOf': [
                                {'type': 'null'},
                                {
                                    '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema.<locals>.InnerData2'
                                },
                            ]
                        }
                    },
                    'type': 'object',
                },
                'tests.json_schema.test_convert.test_to_json_schema.<locals>.AnotherInnerData': {
                    'description': 'AnotherInnerData',
                    'properties': {'bar': {'type': 'string'}},
                    'required': ['bar'],
                    'type': 'object',
                },
            },
        },
    }


@pytest.mark.skipif(sys.version_info < (3, 10), reason='New style unions available after 3.10')
def test_to_json_schema__new_union_syntax():
    @dataclass
    class Data:
        """Docs"""

        a: int | None

    serializer = Serializer(Data)

    assert serializer.get_json_schema() == {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__new_union_syntax.<locals>.Data',
        'components': {
            'schemas': {
                'tests.json_schema.test_convert.test_to_json_schema__new_union_syntax.<locals>.Data': {
                    'description': 'Docs',
                    'properties': {'a': {'anyOf': [{'type': 'null'}, {'type': 'integer'}]}},
                    'required': ['a'],
                    'type': 'object',
                }
            },
        },
    }


@dataclass
class TreeNode:
    """Node"""

    data: str
    left: Optional['TreeNode'] = None
    right: Optional['TreeNode'] = None


def test_to_json_schema__recursive_type():
    serializer = Serializer(TreeNode)

    assert serializer.get_json_schema() == {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        '$ref': '#/components/schemas/tests.json_schema.test_convert.TreeNode',
        'components': {
            'schemas': {
                'tests.json_schema.test_convert.TreeNode': {
                    'description': 'Node',
                    'properties': {
                        'data': {'type': 'string'},
                        'left': {
                            'anyOf': [
                                {'type': 'null'},
                                {'$ref': '#/components/schemas/tests.json_schema.test_convert.TreeNode'},
                            ]
                        },
                        'right': {
                            'anyOf': [
                                {'type': 'null'},
                                {'$ref': '#/components/schemas/tests.json_schema.test_convert.TreeNode'},
                            ]
                        },
                    },
                    'required': ['data'],
                    'type': 'object',
                }
            }
        },
    }


def test_to_json_schema__literal():
    class BarEnum(Enum):
        bar = 'bar'

    serializer = Serializer(Literal['foo', 1, BarEnum.bar])
    assert serializer.get_json_schema() == {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'enum': ['foo', 1, 'bar'],
    }


def test_to_json_schema__tagged_union():
    @dataclass
    class Foo:
        val: int
        type: Literal['foo']

    @dataclass
    class Bar:
        val: str
        type: Literal['bar'] = 'bar'

    class BuzEnum(Enum):
        buz = 'buz'

    @dataclass
    class Baz:
        val: float
        type: Literal[BuzEnum.buz] = BuzEnum.buz

    @dataclass
    class Base:
        field: Annotated[Union[Foo, Bar, Baz], Discriminator('type')]

    serializer = Serializer(Base)
    assert serializer.get_json_schema() == {
        '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Base',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                'tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Bar': {
                    'properties': {'type': {'enum': ['bar']}, 'val': {'type': 'string'}},
                    'required': ['val', 'type'],
                    'type': 'object',
                },
                'tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Base': {
                    'properties': {
                        'field': {
                            'discriminator': {
                                'mapping': {
                                    'bar': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Bar',
                                    'buz': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Baz',
                                    'foo': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Foo',
                                },
                                'propertyName': 'type',
                            },
                            'oneOf': [
                                {
                                    '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Foo'
                                },
                                {
                                    '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Bar'
                                },
                                {
                                    '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Baz'
                                },
                            ],
                        }
                    },
                    'required': ['field'],
                    'type': 'object',
                },
                'tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Baz': {
                    'properties': {'type': {'enum': ['buz']}, 'val': {'type': 'number'}},
                    'required': ['val', 'type'],
                    'type': 'object',
                },
                'tests.json_schema.test_convert.test_to_json_schema__tagged_union.<locals>.Foo': {
                    'properties': {'type': {'enum': ['foo']}, 'val': {'type': 'integer'}},
                    'required': ['val', 'type'],
                    'type': 'object',
                },
            }
        },
    }


def test_to_json_schema__union():
    @dataclass
    class Foo:
        val: int

    @dataclass
    class Bar:
        val: str

    @dataclass
    class Base:
        field: Union[Foo, Bar, str]

    serializer = Serializer(Base)
    assert serializer.get_json_schema() == {
        '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__union.<locals>.Base',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                'tests.json_schema.test_convert.test_to_json_schema__union.<locals>.Bar': {
                    'properties': {'val': {'type': 'string'}},
                    'required': ['val'],
                    'type': 'object',
                },
                'tests.json_schema.test_convert.test_to_json_schema__union.<locals>.Base': {
                    'properties': {
                        'field': {
                            'oneOf': [
                                {
                                    '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__union.<locals>.Foo'
                                },
                                {
                                    '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__union.<locals>.Bar'
                                },
                                {'type': 'string'},
                            ],
                        }
                    },
                    'required': ['field'],
                    'type': 'object',
                },
                'tests.json_schema.test_convert.test_to_json_schema__union.<locals>.Foo': {
                    'properties': {'val': {'type': 'integer'}},
                    'required': ['val'],
                    'type': 'object',
                },
            }
        },
    }


def test_to_json_schema__force_none_as_default_for_optional():
    @dataclass
    class Data:
        a: Optional[int]

    serializer = Serializer(Data, force_default_for_optional=True)
    assert serializer.get_json_schema() == {
        '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__force_none_as_default_for_optional.<locals>.Data',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                'tests.json_schema.test_convert.test_to_json_schema__force_none_as_default_for_optional.<locals>.Data': {
                    'properties': {'a': {'anyOf': [{'type': 'null'}, {'type': 'integer'}]}},
                    'type': 'object',
                }
            }
        },
    }


def test_to_json_schema__bytes():
    @dataclass
    class Data:
        a: bytes

    serializer = Serializer(Data)
    assert serializer.get_json_schema() == {
        '$ref': '#/components/schemas/tests.json_schema.test_convert.test_to_json_schema__bytes.<locals>.Data',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                'tests.json_schema.test_convert.test_to_json_schema__bytes.<locals>.Data': {
                    'properties': {'a': {'type': 'string', 'format': 'binary'}},
                    'type': 'object',
                    'required': ['a'],
                }
            }
        },
    }


def test_enum_x_attrs():
    class EnumCls(Enum):
        a = 'a'
        """a docstring"""

    serializer = Serializer(EnumCls)
    assert serializer.get_json_schema() == {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'enum': ['a'],
        'type': 'string',
        'x-a': 'a docstring',
    }


def test_enum_multiply_types():
    class EnumCls(Enum):
        a = 'a'
        """a docstring"""
        b = 1
        """b docstring"""

    serializer = Serializer(EnumCls)
    assert serializer.get_json_schema() == {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'enum': ['a', 1],
        'x-a': 'a docstring',
        'x-1': 'b docstring',
    }


def test_one_dataclass_with_different_annotations__should_generate_different_schemas():
    @dataclass
    class Foo:
        filed_name: str

    @dataclass
    class Bar:
        foo1: Foo
        foo2: Annotated[Foo, CamelCase]

    schema = Serializer(Bar)
    name_prefix = (
        'tests.json_schema.test_convert.'
        'test_one_dataclass_with_different_annotations__should_generate_different_schemas.<locals>'
    )
    assert schema.get_json_schema() == {
        '$ref': f'#/components/schemas/{name_prefix}.Bar',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                f'{name_prefix}.Bar': {
                    'properties': {
                        'foo1': {'$ref': f'#/components/schemas/{name_prefix}.Foo'},
                        'foo2': {'$ref': f'#/components/schemas/{name_prefix}.Foo1'},
                    },
                    'required': ['foo1', 'foo2'],
                    'type': 'object',
                },
                f'{name_prefix}.Foo': {
                    'properties': {'filed_name': {'type': 'string'}},
                    'required': ['filed_name'],
                    'type': 'object',
                },
                f'{name_prefix}.Foo1': {
                    'properties': {'filedName': {'type': 'string'}},
                    'required': ['filedName'],
                    'type': 'object',
                },
            }
        },
    }


class TestJsonSchemaBuilder:
    def test_build__use_custom_ref_prefix(self):
        @dataclass
        class Data:
            a: int

        serializer = Serializer(Data)
        schema_builder = JsonSchemaBuilder(ref_prefix='#/foo/bar')
        schema = schema_builder.build(serializer)
        definitions = schema_builder.get_definitions()

        assert schema == {
            '$ref': '#/foo/bar/tests.json_schema.test_convert.TestJsonSchemaBuilder.test_build__use_custom_ref_prefix.<locals>.Data',
        }

        assert definitions == {
            'tests.json_schema.test_convert.TestJsonSchemaBuilder.test_build__use_custom_ref_prefix.<locals>.Data': {
                'properties': {'a': {'type': 'integer'}},
                'required': ['a'],
                'type': 'object',
            }
        }

    def test_build__add_dialect_uri(self):
        serializer = Serializer(int)
        schema_builder = JsonSchemaBuilder(add_dialect_uri=True)
        schema = schema_builder.build(serializer)

        assert schema == {
            'type': 'integer',
            '$schema': 'https://json-schema.org/draft/2020-12/schema',
        }
        assert schema_builder.get_definitions() == {}



import pytest

import serpyco_rs

from .compare.libs import serpyco_rs as serializer
from .utils import repeat


def test_dump(bench_or_check_refcount):
    serializer.dump(serializer.test_object)  # warmup
    bench_or_check_refcount.group = 'dump full'
    bench_or_check_refcount(repeat(lambda: serializer.dump(serializer.test_object), count=100))


def test_load_validate(bench_or_check_refcount):
    test_dict = serializer.dump(serializer.test_object)
    serializer.load(test_dict)  # warmup

    bench_or_check_refcount.group = 'load full with validate'
    bench_or_check_refcount(repeat(lambda: serializer.load(test_dict), count=100))


@pytest.mark.skip(reason='Need fix refcount leak')
def test_create_serializer(bench_or_check_refcount):
    bench_or_check_refcount.group = 'create serializer'
    bench_or_check_refcount(repeat(lambda: serpyco_rs.Serializer(serializer.Dataclass), count=100))



import sys
import textwrap
import uuid
from dataclasses import dataclass
from datetime import date, datetime, time
from decimal import Decimal
from enum import Enum
from typing import Annotated, Any, Callable, Literal, Optional, Union
from unittest import mock

import pytest
from serpyco_rs import Serializer
from serpyco_rs.exceptions import ErrorItem, SchemaValidationError, ValidationError
from serpyco_rs.metadata import Discriminator, Max, MaxLength, Min, MinLength
from typing_extensions import NotRequired, Required, TypedDict


class EnumTest(Enum):
    foo = 'foo'
    bar = 'bar'


@dataclass
class EntityTest:
    key: str


@dataclass
class Foo:
    type: Literal['foo']
    val: int


@dataclass
class Bar:
    type: Literal['bar']
    val: str


class TypedDictTotalTrue(TypedDict):
    foo: int
    bar: NotRequired[str]


class TypedDictTotalFalse(TypedDict, total=False):
    foo: int
    bar: Required[str]


@pytest.mark.parametrize(
    ['cls', 'value'],
    (
        (bool, True),
        (bool, False),
        (str, ''),
        (Annotated[str, MinLength(1), MaxLength(3)], '12'),
        (int, -99),
        (Annotated[int, Min(1), Max(1000)], 99),
        (float, 1.3),
        (Annotated[float, Min(0), Max(0.4)], 0.1),
        (Decimal, '0.1'),  # support str
        (Decimal, 0.1),  # or int input
        (Decimal, 'NaN'),  # or int input
        (uuid.UUID, str(uuid.uuid4())),  # support only str input
        (time, '12:34'),
        (time, '12:34:56'),
        (time, '12:34:56.000078'),
        (time, '12:34Z'),
        (time, '12:34+0300'),
        (time, '12:34+03:00'),
        (time, '12:34:00+03:00'),
        (time, '12:34:56.000078+03:00'),
        (time, '12:34:56.000078+00:00'),
        (datetime, '2022-10-10T14:23:43'),
        (datetime, '2022-10-10T14:23:43.123456'),
        (datetime, '2022-10-10T14:23:43.123456Z'),
        (datetime, '2022-10-10T14:23:43.123456+00:00'),
        (datetime, '2022-10-10T14:23:43.123456-00:00'),
        (date, '2020-07-17'),
        (EnumTest, 'foo'),
        (Optional[int], None),
        (Optional[int], 1),
        (EntityTest, {'key': 'val'}),
        (list[int], [1, 2]),
        (dict[str, int], {'a': 1}),
        (tuple[str, int, bool], ['1', 2, True]),
        (Annotated[Union[Foo, Bar], Discriminator('type')], {'type': 'foo', 'val': 1}),
        (Annotated[Union[Foo, Bar], Discriminator('type')], {'type': 'bar', 'val': '1'}),
        (Any, ['1', 2, True]),
        (Any, {}),
        (TypedDictTotalTrue, {'foo': 1}),
        (TypedDictTotalTrue, {'foo': 1, 'bar': '1'}),
        (TypedDictTotalFalse, {'bar': '1'}),
        (TypedDictTotalFalse, {'foo': 1, 'bar': '1'}),
    ),
)
def test_validate(cls, value):
    Serializer(cls).load(value)


if sys.version_info >= (3, 10):

    @pytest.mark.parametrize(
        ['cls', 'value'],
        (
            (Optional[int], None),
            (int | None, None),
            (int | None, 2),
        ),
    )
    def test_validate_new_union(cls, value):
        Serializer(cls).load(value)


def _mk_e(m=mock.ANY, ip=mock.ANY) -> Callable[[ErrorItem], None]:
    def cmp(e: ErrorItem) -> None:
        assert e.message == m and e.instance_path == ip

    return cmp


@pytest.mark.parametrize(
    ['cls', 'value', 'check'],
    (
        (bool, 1, _mk_e(m='1 is not of type "boolean"')),
        (str, 1, _mk_e(m='1 is not of type "string"')),
        (
            Annotated[str, MinLength(2)],
            'a',
            _mk_e(m='"a" is shorter than 2 characters'),
        ),
        (
            Annotated[str, MaxLength(2)],
            'aaa',
            _mk_e(m='"aaa" is longer than 2 characters'),
        ),
        (int, 9.1, _mk_e(m='9.1 is not of type "integer"')),
        (int, '9', _mk_e(m='"9" is not of type "integer"')),
        (
            Annotated[int, Min(1)],
            0,
            _mk_e(m='0 is less than the minimum of 1'),
        ),
        (
            Annotated[int, Max(1)],
            10,
            _mk_e(m='10 is greater than the maximum of 1'),
        ),
        (float, None, _mk_e(m='None is not of type "number"')),
        (
            Annotated[float, Min(1.1)],
            0.1,
            _mk_e(m='0.1 is less than the minimum of 1.1'),
        ),
        (
            Annotated[float, Max(1.5)],
            10.1,
            _mk_e(m='10.1 is greater than the maximum of 1.5'),
        ),
        (uuid.UUID, 'asd', _mk_e(ip='')),
        (time, '12:34:a', _mk_e(ip='')),
        (datetime, '2022-10-10//12', _mk_e(ip='')),
        (date, '17-02-2022', _mk_e(ip='')),
        (datetime, '2022-10-10T14:23:43.123456-30:00', _mk_e(ip='')),
        (EnumTest, 'buz', _mk_e(m='"buz" is not one of ["foo", "bar"]')),
        (
            Optional[int],
            'foo',
            _mk_e(m='"foo" is not of type "integer"'),
        ),
        (EntityTest, {}, _mk_e(m='"key" is a required property')),
        (
            list[int],
            [1, '1'],
            _mk_e(m='"1" is not of type "integer"', ip='1'),
        ),
        (
            dict[str, int],
            {'a': '1'},
            _mk_e(m='"1" is not of type "integer"', ip='a'),
        ),
        (
            tuple[str, int, bool],
            ['1'],
            _mk_e(m="['1'] has less than 3 items"),
        ),
        (
            tuple[str, int, bool],
            ['1', 1, True, 0],
            _mk_e(m="['1', 1, True, 0] has more than 3 items"),
        ),
        (
            tuple[str, bool],
            ['1', 1],
            _mk_e(
                m='1 is not of type "boolean"',
                ip='1',
            ),
        ),
        (
            Annotated[Union[Foo, Bar], Discriminator('type')],
            {'type': 'buz'},
            _mk_e(m='"buz" is not one of ["foo", "bar"] discriminator values'),
        ),
        (
            Annotated[Union[Foo, Bar], Discriminator('type')],
            {'type': 'foo', 'val': '123'},
            _mk_e(ip='val', m='"123" is not of type "integer"'),
        ),
        (
            Annotated[Union[Foo, Bar], Discriminator('type')],
            {'type': 'bar', 'val': 1},
            _mk_e(ip='val', m='1 is not of type "string"'),
        ),
        (TypedDictTotalTrue, {}, _mk_e(m='"foo" is a required property')),
        (TypedDictTotalFalse, {}, _mk_e(m='"bar" is a required property')),
    ),
)
def test_validate__validation_error(cls, value, check):
    serializer = Serializer(cls)
    with pytest.raises(SchemaValidationError) as exc_info:
        serializer.load(value)

    assert len(exc_info.value.errors) == 1
    check(exc_info.value.errors[0])


def test_validate__error_format():
    @dataclass
    class Inner:
        baz: str

    @dataclass
    class A:
        foo: int
        bar: Inner

    serializer = Serializer(A)

    value = {'foo': '1', 'bar': {'buz': None}, 'qux': 0}

    with pytest.raises(SchemaValidationError) as exc_info:
        serializer.load(value)

    assert exc_info.value.errors == [
        ErrorItem(
            message='"1" is not of type "integer"',
            instance_path='foo',
        ),
    ]


def test_validation_exceptions_inheritance():
    serializer = Serializer(int)
    with pytest.raises(SchemaValidationError) as exc_info:
        serializer.load('1')

    assert isinstance(exc_info.value, ValidationError)
    assert exc_info.value.message == 'Schema validation failed'


def test_validation_error_message():
    @dataclass
    class A:
        foo: int
        bar: str

    serializer = Serializer(A)
    with pytest.raises(SchemaValidationError) as exc_info:
        serializer.load({'foo': '1', 'bar': 2})

    error = exc_info.value
    error_item = error.errors[0]

    assert str(error_item) == (""""1" is not of type "integer" (instance_path='foo')""")
    assert repr(error_item) == ("""ErrorItem(message='"1" is not of type "integer"', instance_path='foo')""")
    assert str(error) == textwrap.dedent(
        """\
    Schema validation failed:
    - "1" is not of type "integer" (instance_path='foo')
      """
    )
    assert repr(error) == textwrap.dedent(
        """\
    SchemaValidationError(
        message="Schema validation failed",
        errors=[
            ErrorItem(message='"1" is not of type "integer"', instance_path='foo'),
        ]
    )"""
    )



from dataclasses import dataclass
from typing import TypeVar, Generic, List, Type

from serpyco_rs._type_utils import get_type_hints

T = TypeVar('T')
U = TypeVar('U')
KT = TypeVar('KT')
VT = TypeVar('VT')


def test_get_type_hints():
    @dataclass
    class A:
        a: int
        b: str

    assert get_type_hints(A) == {'a': int, 'b': str}


def test_get_type_hints_with_inheritance():
    @dataclass
    class A:
        a: int
        b: str

    @dataclass
    class B(A):
        c: float

    assert get_type_hints(B) == {'a': int, 'b': str, 'c': float}


def test_get_type_hints_with_nested_dataclasses():
    @dataclass
    class A:
        a: int
        b: str

    @dataclass
    class B:
        a: A
        c: float

    assert get_type_hints(B) == {'a': A, 'c': float}


def test_get_type_hints_with_generics():
    @dataclass
    class A(Generic[T, U]):
        a: T
        b: U

    @dataclass
    class B(A[int, U]):
        pass

    assert get_type_hints(B[bool]) == {'a': int, 'b': bool}


def test_get_type_hints_with_generics_and_reuse_type_vars():
    @dataclass
    class A(Generic[T, U]):
        a: T
        b: U

    @dataclass
    class B(A[int, T]):
        pass

    assert get_type_hints(B[str]) == {'a': int, 'b': str}


def test_get_type_hints_with_generics_and_nested_dataclasses():
    @dataclass
    class A(Generic[T, U]):
        a: T
        b: U

    @dataclass
    class B:
        a: A[int, str]
        c: float

    assert get_type_hints(B) == {'a': A[int, str], 'c': float}


def test_get_type_hints_with_generics_and_annotations():
    @dataclass
    class A(Generic[T, U]):
        a: T
        b: U

    @dataclass
    class B:
        f: A[int, str]
        g: A[float, str]

    assert get_type_hints(B) == {'f': A[int, str], 'g': A[float, str]}


def test_get_type_hints_generic_alias():
    class Foo(Generic[T, KT]):
        arr: List[T]
        key: KT

    assert get_type_hints(Foo[str, int]) == {'arr': List[str], 'key': int}


def test_get_type_hints_partial_typevars():
    class Foo(Generic[T, KT]):
        arr: List[T]
        key: KT

    class Bar(Foo[str, KT]):
        pass

    assert get_type_hints(Bar) == {'arr': List[str], 'key': KT}
    assert get_type_hints(Bar[bool]) == {'arr': List[str], 'key': bool}


def test_get_type_hints_swapped_typevar_order():
    class Foo(Generic[T, KT]):
        arr: List[T]
        key: KT

    class Bar(Foo[KT, T]):
        pass

    assert get_type_hints(Bar) == {'arr': List[KT], 'key': T}
    assert get_type_hints(Bar) == {'arr': List[KT], 'key': T}


def test_get_type_hints_multiple_generic_inheritance():
    class Foo(Generic[T, KT]):
        arr: List[T]
        key: KT

    class Bar(Generic[T]):
        beer: T

    class Baz(Foo[T, KT], Bar[T], Generic[T, KT]):
        pass

    class Baz2(Foo[T, KT], Bar[T]):
        pass

    assert get_type_hints(Baz[int, bool]) == {'arr': List[int], 'key': bool, 'beer': int}
    assert get_type_hints(Baz2[int, bool]) == {'arr': List[int], 'key': bool, 'beer': int}


def test_get_type_hints_unorthodox_generic_placement():
    class A(Generic[T, VT]):
        a_type: T
        value: VT

    class B(Generic[KT, T]):
        b_type: T
        key: KT

    class C(A[T, VT], Generic[VT, T, KT], B[KT, T]):
        pass

    assert get_type_hints(C) == {'a_type': T, 'value': VT, 'b_type': T, 'key': KT}


def test_get_type_hints_extended_generic_rules_subclassing():
    class T1(Type[T]):
        value: T

    class T2(list[T]):
        value: T

    assert get_type_hints(T1) == {'value': T}
    assert get_type_hints(T2) == {'value': T}
    assert get_type_hints(T1[bool]) == {'value': bool}


def test_get_type_hints_generic_class_no_params_required():
    class Foo(Generic[T, KT]):
        arr: List[T]
        key: KT

    class Bar(Foo[str, int]):
        pass

    assert get_type_hints(Bar) == {'arr': List[str], 'key': int}


def test_get_type_hints_unbound_typevar():
    # https://github.com/python/cpython/pull/111515#issuecomment-2018336687

    class Foo(Generic[T]):
        x: list[T]
        y: KT

    assert get_type_hints(Foo) == {'x': list[T], 'y': KT}  # ok
    assert get_type_hints(Foo[int]) == {'x': list[int], 'y': KT}  # error



import sys
import typing
from enum import Enum
from functools import singledispatch
from typing import Any, Optional, Union, cast

from .._utils import get_attributes_doc
from ._consts import DEFAULT_REF_PREFIX, SCHEMA_VERSION


if sys.version_info >= (3, 10):
    from typing import TypeGuard
else:
    from typing_extensions import TypeGuard

from .. import _describe as describe
from ._entities import (
    ArrayType,
    Boolean,
    Config,
    DiscriminatedUnionType,
    Discriminator,
    IntegerType,
    Null,
    NumberType,
    ObjectType,
    RefType,
    Schema,
    StringType,
)


if typing.TYPE_CHECKING:
    from .._main import Serializer


class JsonSchemaBuilder:
    def __init__(self, add_dialect_uri: bool = False, ref_prefix: str = DEFAULT_REF_PREFIX):
        self._definitions: dict[str, Any] = {}
        self._ref_prefix = ref_prefix.rstrip('/')
        self._add_dialect_uri = add_dialect_uri
        self._config = Config(ref_prefix=ref_prefix)

    def build(self, serializer: 'Serializer[Any]') -> dict[str, Any]:
        schema = to_json_schema(serializer._type_info, config=self._config)
        schema_def = schema.dump(self._definitions)
        if self._add_dialect_uri:
            schema_def['$schema'] = SCHEMA_VERSION
        return schema_def

    def get_definitions(self) -> dict[str, Any]:
        return self._definitions


def get_json_schema(t: describe.BaseType) -> dict[str, Any]:
    schema = to_json_schema(t, config=Config())
    definitions: dict[str, Any] = {}
    schema_def = schema.dump(definitions)
    components = {'components': {'schemas': definitions}} if definitions else {}
    return {
        '$schema': SCHEMA_VERSION,
        **schema_def,
        **components,
    }


@singledispatch
def to_json_schema(_: Any, doc: Optional[str] = None, *, config: Config) -> Schema:
    return Schema(description=doc, config=config)


@to_json_schema.register
def _(arg: describe.StringType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return StringType(minLength=arg.min_length, maxLength=arg.max_length, description=doc, config=config)


@to_json_schema.register
def _(arg: describe.IntegerType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return IntegerType(minimum=arg.min, maximum=arg.max, description=doc, config=config)


@to_json_schema.register
def _(_: describe.BytesType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return StringType(format='binary', description=doc, config=config)


@to_json_schema.register
def _(arg: describe.FloatType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return NumberType(minimum=arg.min, maximum=arg.max, description=doc, config=config)


@to_json_schema.register
def _(_: describe.DecimalType, doc: Optional[str] = None, *, config: Config) -> Schema:
    # todo: support min/max
    return Schema(
        oneOf=[
            StringType(format='decimal', config=config),
            NumberType(format='decimal', config=config),
        ],
        description=doc,
        config=config,
    )


@to_json_schema.register
def _(_: describe.BooleanType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return Boolean(config=config, description=doc)


@to_json_schema.register
def _(_: describe.UUIDType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return StringType(format='uuid', description=doc, config=config)


@to_json_schema.register
def _(_: describe.TimeType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return StringType(format='time', description=doc, config=config)


@to_json_schema.register
def _(_: describe.DateTimeType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return StringType(format='date-time', description=doc, config=config)


@to_json_schema.register
def _(_: describe.DateType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return StringType(format='date', description=doc, config=config)


@to_json_schema.register
def _(arg: describe.EnumType, doc: Optional[str] = None, *, config: Config) -> Schema:
    docs = get_attributes_doc(arg.cls)
    enum_values = [item.value for item in arg.items]
    type_ = None
    if (types := {type(item.value) for item in arg.items}) and len(types) == 1:
        type_ = {int: 'integer', str: 'string'}.get(types.pop(), None)

    return Schema(
        type=type_,
        enum=enum_values,
        description=doc,
        config=config,
        additionalArgs={f'x-{item.value}': docs.get(item.name) for item in arg.items},
    )


@to_json_schema.register
def _(arg: describe.OptionalType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return Schema(
        anyOf=[
            Null(config=config),
            to_json_schema(arg.inner, config=config),
        ],
        description=doc,
        config=config,
    )


@to_json_schema.register
def _(arg: describe.EntityType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return ObjectType(
        properties={prop.dict_key: to_json_schema(prop.field_type, prop.doc, config=config) for prop in arg.fields},
        required=[prop.dict_key for prop in arg.fields if prop.required] or None,
        name=arg.name,
        description=arg.doc,
        config=config,
    )


@to_json_schema.register
def _(arg: describe.TypedDictType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return ObjectType(
        properties={prop.dict_key: to_json_schema(prop.field_type, prop.doc, config=config) for prop in arg.fields},
        required=[prop.dict_key for prop in arg.fields if prop.required] or None,
        name=arg.name,
        description=arg.doc,
        config=config,
    )


@to_json_schema.register
def _(arg: describe.ArrayType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return ArrayType(items=to_json_schema(arg.item_type, config=config), description=doc, config=config)


@to_json_schema.register
def _(arg: describe.DictionaryType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return ObjectType(
        additionalProperties=to_json_schema(arg.value_type, config=config), description=doc, config=config
    )


@to_json_schema.register
def _(arg: describe.TupleType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return ArrayType(
        prefixItems=[to_json_schema(item, config=config) for item in arg.item_types],
        minItems=len(arg.item_types),
        maxItems=len(arg.item_types),
        description=doc,
        config=config,
    )


@to_json_schema.register
def _(_: describe.AnyType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return Schema(description=doc, config=config)


@to_json_schema.register
def _(holder: describe.RecursionHolder, doc: Optional[str] = None, *, config: Config) -> Schema:
    return RefType(description=doc, name=holder.name, config=config)


@to_json_schema.register
def _(arg: describe.LiteralType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return Schema(
        enum=[arg.value if isinstance(arg, Enum) else arg for arg in arg.args],
        description=doc,
        config=config,
    )


@to_json_schema.register
def _(arg: describe.UnionType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return Schema(
        oneOf=[to_json_schema(t, config=config) for t in arg.item_types],
        description=doc,
        config=config,
    )


@to_json_schema.register
def _(arg: describe.DiscriminatedUnionType, doc: Optional[str] = None, *, config: Config) -> Schema:
    objects = {
        name: schema
        for name, t in arg.item_types.items()
        if (schema := to_json_schema(t, config=config)) and _check_unions_schema_types(schema)
    }

    return DiscriminatedUnionType(
        oneOf=list(objects.values()),
        discriminator=Discriminator(
            property_name=arg.load_discriminator,
            mapping={name: val.ref for name, val in objects.items()},
        ),
        description=doc,
        config=config,
    )


def _check_unions_schema_types(schema: Schema) -> TypeGuard[Union[ObjectType, RefType]]:
    if isinstance(schema, (ObjectType, RefType)):
        return True
    raise RuntimeError(f'Unions schema items must be ObjectType or RefType. Current: {schema}')


@to_json_schema.register
def _(arg: describe.CustomType, doc: Optional[str] = None, *, config: Config) -> Schema:
    return Schema(additionalArgs=arg.json_schema, description=doc, config=config)



from collections.abc import Sequence
from enum import Enum, IntEnum
from typing import Any, Callable, Generic, TypeVar

from ._meta import Meta, MetaStateKey

_T = TypeVar('_T')
_I = TypeVar('_I')
_O = TypeVar('_O')

class ValidationError(Exception):
    message: str

class ErrorItem:
    message: str
    instance_path: str

    def __init__(self, message: str, instance_path: str): ...

class SchemaValidationError(ValidationError):
    errors: list[ErrorItem]

class Serializer(Generic[_T]):
    def __init__(self, py_class: BaseType, naive_datetime_to_utc: bool): ...
    def dump(self, value: _T) -> Any: ...
    def load(self, data: Any) -> _T: ...
    def load_query_params(self, data: Any) -> _T: ...

class CustomEncoder(Generic[_I, _O]):
    serialize: Callable[[_I], _O] | None
    deserialize: Callable[[_O], _I] | None

    def __init__(self, serialize: Callable[[_I], _O] | None = None, deserialize: Callable[[_O], _I] | None = None): ...

class BaseType:
    custom_encoder: CustomEncoder[Any, Any] | None

    def __init__(self, custom_encoder: CustomEncoder[Any, Any] | None): ...

class IntegerType(BaseType):
    min: int | None
    max: int | None

    def __init__(
        self, min: int | None = None, max: int | None = None, custom_encoder: CustomEncoder[Any, Any] | None = None
    ): ...

class FloatType(BaseType):
    min: float | None
    max: float | None

    def __init__(self, min: float | None, max: float | None, custom_encoder: CustomEncoder[Any, Any] | None): ...

class DecimalType(BaseType):
    min: float | None
    max: float | None

    def __init__(self, min: float | None, max: float | None, custom_encoder: CustomEncoder[Any, Any] | None): ...

class StringType(BaseType):
    min_length: int | None
    max_length: int | None

    def __init__(
        self,
        min_length: int | None = None,
        max_length: int | None = None,
        custom_encoder: CustomEncoder[Any, Any] | None = None,
    ): ...

class BooleanType(BaseType):
    def __init__(self, custom_encoder: CustomEncoder[Any, Any] | None): ...

class UUIDType(BaseType):
    def __init__(self, custom_encoder: CustomEncoder[Any, Any] | None): ...

class TimeType(BaseType):
    def __init__(self, custom_encoder: CustomEncoder[Any, Any] | None): ...

class DateTimeType(BaseType):
    def __init__(self, custom_encoder: CustomEncoder[Any, Any] | None): ...

class DateType(BaseType):
    def __init__(self, custom_encoder: CustomEncoder[Any, Any] | None): ...

class DefaultValue(Generic[_T]):
    @staticmethod
    def none() -> DefaultValue[None]: ...
    @staticmethod
    def some(value: _T) -> DefaultValue[_T]: ...
    def is_none(self) -> bool: ...

NOT_SET: DefaultValue[None]

class EntityField(BaseType):
    name: str
    dict_key: str
    field_type: BaseType
    required: bool = True
    is_discriminator_field: bool = False
    default: DefaultValue[Any]
    default_factory: DefaultValue[Callable[[], Any]]
    doc: str | None

    def __init__(
        self,
        name: str,
        dict_key: str,
        field_type: BaseType,
        required: bool = True,
        is_discriminator_field: bool = False,
        default: DefaultValue[Any] = ...,
        default_factory: DefaultValue[Callable[[], Any]] | DefaultValue[None] = ...,
        doc: str | None = None,
    ): ...

class EntityType(BaseType):
    cls: type[Any]
    name: str
    fields: Sequence[EntityField]
    omit_none: bool
    is_frozen: bool
    doc: str | None

    def __init__(
        self,
        cls: type[Any],
        name: str,
        fields: Sequence[EntityField],
        omit_none: bool = False,
        is_frozen: bool = False,
        doc: str | None = None,
        custom_encoder: CustomEncoder[Any, Any] | None = None,
    ): ...

class TypedDictType(BaseType):
    name: str
    fields: Sequence[EntityField]
    omit_none: bool
    doc: str | None

    def __init__(
        self,
        name: str,
        fields: Sequence[EntityField],
        omit_none: bool = False,
        doc: str | None = None,
        custom_encoder: CustomEncoder[Any, Any] | None = None,
    ): ...

class ArrayType(BaseType):
    item_type: BaseType

    def __init__(self, item_type: BaseType, custom_encoder: CustomEncoder[Any, Any] | None = None): ...

class EnumType(BaseType):
    cls: type[Enum | IntEnum]
    items: list[Any]

    def __init__(
        self, cls: type[Enum | IntEnum], items: list[Any], custom_encoder: CustomEncoder[Any, Any] | None = None
    ): ...

class OptionalType(BaseType):
    inner: BaseType

    def __init__(self, inner: BaseType, custom_encoder: CustomEncoder[Any, Any] | None = None): ...

class DictionaryType(BaseType):
    key_type: BaseType
    value_type: BaseType
    omit_none: bool

    def __init__(
        self,
        key_type: BaseType,
        value_type: BaseType,
        omit_none: bool = False,
        custom_encoder: CustomEncoder[Any, Any] | None = None,
    ): ...

class TupleType(BaseType):
    item_types: list[BaseType]

    def __init__(self, item_types: list[BaseType], custom_encoder: CustomEncoder[Any, Any] | None = None): ...

class BytesType(BaseType):
    def __init__(self, custom_encoder: CustomEncoder[Any, Any] | None = None): ...

class AnyType(BaseType):
    def __init__(self, custom_encoder: CustomEncoder[Any, Any] | None = None): ...

class UnionType(BaseType):
    item_types: list[BaseType]
    union_repr: str

    def __init__(
        self,
        item_types: list[BaseType],
        union_repr: str,
        custom_encoder: CustomEncoder[Any, Any] | None = None,
    ): ...

class DiscriminatedUnionType(BaseType):
    item_types: dict[str, BaseType]
    dump_discriminator: str
    load_discriminator: str

    def __init__(
        self,
        item_types: dict[str, BaseType],
        dump_discriminator: str,
        load_discriminator: str,
        custom_encoder: CustomEncoder[Any, Any] | None = None,
    ): ...

class LiteralType(BaseType):
    args: list[str | int | Enum]

    def __init__(self, args: list[str | int | Enum], custom_encoder: CustomEncoder[Any, Any] | None = None): ...

class RecursionHolder(BaseType):
    name: str
    state_key: MetaStateKey
    meta: Meta

    def __init__(
        self, name: str, state_key: MetaStateKey, meta: Meta, custom_encoder: CustomEncoder[Any, Any] | None = None
    ): ...
    def get_type(self) -> BaseType: ...

class CustomType(BaseType):
    json_schema: dict[str, Any]

    def __init__(self, custom_encoder: CustomEncoder[Any, Any], json_schema: dict[str, Any]): ...



import dataclasses
import sys
from collections.abc import Callable, Iterable, Mapping, Sequence
from datetime import date, datetime, time
from decimal import Decimal
from enum import Enum, IntEnum
from functools import cache
from typing import (
    Annotated,
    Any,
    ForwardRef,
    Generic,
    Literal,
    Optional,
    TypeVar,
    Union,
    cast,
    get_origin,
    overload,
)
from uuid import UUID

from typing_extensions import NotRequired, Required, assert_never, get_args, is_typeddict

from ._custom_types import CustomType as CustomTypeMeta
from ._impl import (
    NOT_SET,
    AnyType,
    ArrayType,
    BaseType,
    BooleanType,
    BytesType,
    CustomEncoder,
    CustomType,
    DateTimeType,
    DateType,
    DecimalType,
    DefaultValue,
    DictionaryType,
    DiscriminatedUnionType,
    EntityField,
    EntityType,
    EnumType,
    FloatType,
    IntegerType,
    LiteralType,
    OptionalType,
    RecursionHolder,
    StringType,
    TimeType,
    TupleType,
    TypedDictType,
    UnionType,
    UUIDType,
)
from ._meta import Meta, MetaStateKey
from ._type_utils import get_type_hints  # type: ignore[attr-defined]
from ._utils import get_attributes_doc, to_camelcase
from .metadata import (
    Alias,
    Discriminator,
    FieldFormat,
    Format,
    KeepDefaultForOptional,
    KeepNone,
    Max,
    MaxLength,
    Min,
    MinLength,
    NoFormat,
    NoneAsDefaultForOptional,
    NoneFormat,
    OmitNone,
)


if sys.version_info >= (3, 10):  # pragma: no cover
    from types import UnionType as StdUnionType
else:  # pragma: no cover
    StdUnionType = None

FrozenInstanceErrors: tuple[type[Exception], ...] = (dataclasses.FrozenInstanceError,)

try:
    import attr
    from attr.exceptions import FrozenInstanceError

    FrozenInstanceErrors += (FrozenInstanceError,)
except ImportError:  # pragma: no cover
    attr = None  # type: ignore
    FrozenInstanceError = None  # type: ignore


_NoneType = type(None)

_T = TypeVar('_T')


def describe_type(
    t: Any,
    meta: Optional[Meta] = None,
    custom_type_resolver: Optional[Callable[[Any], Optional[CustomTypeMeta[Any, Any]]]] = None,
) -> BaseType:
    args: tuple[Any, ...] = ()
    metadata = _get_annotated_metadata(t)
    type_repr = repr(t)
    if get_origin(t) == Annotated:  # unwrap annotated
        t = t.__origin__
    original_t = t
    if get_origin(t) in {Required, NotRequired}:  # unwrap TypedDict special forms
        t = t.__args__[0]
    if hasattr(t, '__origin__'):
        args = t.__args__
        t = t.__origin__
    # StdUnionType has no __origin__
    elif StdUnionType and isinstance(t, StdUnionType):  # type: ignore[truthy-function]
        args = t.__args__
        t = Union
    elif hasattr(t, '__parameters__'):
        # Если передан generic-класс без type-параметров, значит по PEP-484 заменяем все параметры на Any
        args = (Any,) * len(t.__parameters__)

    if not meta:
        meta = Meta(globals=_get_globals(t), state={})

    t = _evaluate_forwardref(t, meta)

    filed_format = _find_metadata(metadata, FieldFormat, NoFormat)
    none_format = _find_metadata(metadata, NoneFormat, KeepNone)
    none_as_default_for_optional = _find_metadata(metadata, NoneAsDefaultForOptional, KeepDefaultForOptional)
    custom_encoder = _find_metadata(metadata, CustomEncoder)
    annotation_wrapper = _wrap_annotated([filed_format, none_format, none_as_default_for_optional])

    meta_key = MetaStateKey(
        cls=original_t,
        field_format=filed_format,
        none_format=none_format,
        none_as_default_for_optional=none_as_default_for_optional,
    )

    if meta.has_in_state(meta_key):
        return RecursionHolder(
            name=_generate_name(original_t, filed_format, none_format, none_as_default_for_optional),
            state_key=meta_key,
            meta=meta,
            custom_encoder=None,
        )

    if custom_type_resolver and (custom_type := custom_type_resolver(t)):
        if custom_encoder is None:
            custom_encoder = CustomEncoder(
                serialize=custom_type.serialize,
                deserialize=custom_type.deserialize,
            )
        return CustomType(custom_encoder=custom_encoder, json_schema=custom_type.get_json_schema())

    if t is Any:
        return AnyType(custom_encoder=custom_encoder)

    if isinstance(t, type):
        simple_type_mapping: Mapping[type, type[BaseType]] = {
            bytes: BytesType,
            bool: BooleanType,
            date: DateType,
            time: TimeType,
            datetime: DateTimeType,
            UUID: UUIDType,
        }

        if simple := simple_type_mapping.get(t):
            return simple(custom_encoder=custom_encoder)

        number_type_mapping: Mapping[type, type[Union[IntegerType, FloatType]]] = {
            int: IntegerType,
            float: FloatType,
        }

        if number_type := number_type_mapping.get(t):
            min_meta = _find_metadata(metadata, Min)
            max_meta = _find_metadata(metadata, Max)
            return number_type(
                min=cast(Any, min_meta.value) if min_meta else None,
                max=cast(Any, max_meta.value) if max_meta else None,
                custom_encoder=custom_encoder,
            )

        if t is Decimal:
            min_meta = _find_metadata(metadata, Min)
            max_meta = _find_metadata(metadata, Max)
            return DecimalType(
                min=min_meta.value if min_meta else None,
                max=max_meta.value if max_meta else None,
                custom_encoder=custom_encoder,
            )

        if t is str:
            min_length_meta = _find_metadata(metadata, MinLength)
            max_length_meta = _find_metadata(metadata, MaxLength)
            return StringType(
                min_length=min_length_meta.value if min_length_meta else None,
                max_length=max_length_meta.value if max_length_meta else None,
                custom_encoder=custom_encoder,
            )

        if t in {Sequence, list}:
            return ArrayType(
                item_type=(
                    describe_type(annotation_wrapper(args[0]), meta, custom_type_resolver)
                    if args
                    else AnyType(custom_encoder=None)
                ),
                custom_encoder=custom_encoder,
            )

        if t in {Mapping, dict}:
            return DictionaryType(
                key_type=(
                    describe_type(annotation_wrapper(args[0]), meta, custom_type_resolver)
                    if args
                    else AnyType(custom_encoder=None)
                ),
                value_type=(
                    describe_type(annotation_wrapper(args[1]), meta, custom_type_resolver)
                    if args
                    else AnyType(custom_encoder=None)
                ),
                omit_none=none_format.omit,
                custom_encoder=custom_encoder,
            )

        if t is tuple:
            if not args or Ellipsis in args:
                raise RuntimeError('Variable length tuples are not supported')
            return TupleType(
                item_types=[describe_type(annotation_wrapper(arg), meta, custom_type_resolver) for arg in args],
                custom_encoder=custom_encoder,
            )

        if issubclass(t, (Enum, IntEnum)):
            return EnumType(cls=t, items=list(t), custom_encoder=custom_encoder)

        if dataclasses.is_dataclass(t) or _is_attrs(t) or is_typeddict(t):
            meta.add_to_state(meta_key, None)
            entity_type = _describe_entity(
                t=t,
                original_t=original_t,
                cls_filed_format=filed_format,
                cls_none_format=none_format,
                custom_encoder=custom_encoder,
                cls_none_as_default_for_optional=none_as_default_for_optional,
                meta=meta,
                custom_type_resolver=custom_type_resolver,
            )
            meta.add_to_state(meta_key, entity_type)
            return entity_type

    if _is_literal_type(t):
        if args and all(isinstance(arg, (str, int, Enum)) for arg in args):
            return LiteralType(args=list(args), custom_encoder=custom_encoder)
        raise RuntimeError('Supported only Literal[str | int, ...]')

    if _is_new_type(t):
        return describe_type(t.__supertype__, meta, custom_type_resolver)

    if t is Union:
        if _NoneType in args:
            new_args = tuple(arg for arg in args if arg is not _NoneType)
            new_t = Union[new_args] if len(new_args) > 1 else new_args[0]  # type: ignore[unused-ignore]
            return OptionalType(
                inner=describe_type(annotation_wrapper(new_t), meta, custom_type_resolver),
                custom_encoder=None,
            )

        discriminator = _find_metadata(metadata, Discriminator)
        if not discriminator:
            return UnionType(
                item_types=[describe_type(annotation_wrapper(arg), meta, custom_type_resolver) for arg in args],
                union_repr=type_repr.removeprefix('typing.'),
                custom_encoder=custom_encoder,
            )

        if not all(dataclasses.is_dataclass(arg) or _is_attrs(arg) for arg in args):
            raise RuntimeError(
                f'Unions supported only for dataclasses or attrs. Provided: {t}[{",".join(map(str, args))}]'
            )

        meta = dataclasses.replace(meta, discriminator_field=discriminator.name)
        return DiscriminatedUnionType(
            item_types={
                _get_discriminator_value(arg, discriminator.name): describe_type(
                    annotation_wrapper(arg), meta, custom_type_resolver
                )
                for arg in args
            },
            dump_discriminator=discriminator.name,
            load_discriminator=_apply_format(filed_format, discriminator.name),
            custom_encoder=custom_encoder,
        )

    if isinstance(t, TypeVar):
        raise RuntimeError(f'Unfilled TypeVar: {t}')

    raise RuntimeError(f'Unknown type {t!r}')


@dataclasses.dataclass
class _Field(Generic[_T]):
    name: str
    type: type[_T]
    default: Union[DefaultValue[_T], DefaultValue[None]] = NOT_SET
    default_factory: Union[DefaultValue[Callable[[], _T]], DefaultValue[None]] = NOT_SET


def _describe_entity(
    t: Any,
    original_t: Any,
    cls_filed_format: FieldFormat,
    cls_none_format: NoneFormat,
    cls_none_as_default_for_optional: NoneAsDefaultForOptional,
    custom_encoder: Optional[CustomEncoder[Any, Any]],
    meta: Meta,
    custom_type_resolver: Optional[Callable[[Any], Optional[CustomTypeMeta[Any, Any]]]],
) -> Union[EntityType, TypedDictType]:
    # PEP-484: Replace all unfilled type parameters with Any
    if not hasattr(original_t, '__origin__') and getattr(original_t, '__parameters__', None):
        original_t = original_t[(Any,) * len(t.__parameters__)]

    docs = get_attributes_doc(t)
    try:
        types = get_type_hints(original_t, include_extras=True)
    except Exception:  # pylint: disable=broad-except
        types = {}

    discriminator_field = meta.discriminator_field
    meta = dataclasses.replace(meta, globals=_get_globals(t), discriminator_field=None)

    fields = []
    for field in _get_entity_fields(t):
        type_ = types.get(field.name, field.type)
        type_ = Annotated[type_, cls_filed_format, cls_none_format, cls_none_as_default_for_optional]

        metadata = _get_annotated_metadata(type_)
        field_type = describe_type(type_, meta, custom_type_resolver)
        alias = _find_metadata(metadata, Alias)
        none_as_default_for_optional = _find_metadata(metadata, NoneAsDefaultForOptional)

        is_discriminator_field = field.name == discriminator_field
        required = not (field.default != NOT_SET or field.default_factory != NOT_SET) or is_discriminator_field

        default = field.default
        if (
            isinstance(field_type, OptionalType)
            and required
            and none_as_default_for_optional
            and none_as_default_for_optional.use
        ):
            default = DefaultValue.some(None)
            required = False

        fields.append(
            EntityField(
                name=field.name,
                dict_key=alias.value if alias else _apply_format(cls_filed_format, field.name),
                doc=docs.get(field.name),
                field_type=field_type,
                default=default,
                default_factory=field.default_factory,
                is_discriminator_field=is_discriminator_field,
                required=required,
            )
        )

    if is_typeddict(t):
        return TypedDictType(
            name=_generate_name(original_t, cls_filed_format, cls_none_format, cls_none_as_default_for_optional),
            fields=fields,
            omit_none=cls_none_format is OmitNone,
            doc=t.__doc__,
            custom_encoder=custom_encoder,
        )

    return EntityType(
        cls=t,
        name=_generate_name(original_t, cls_filed_format, cls_none_format, cls_none_as_default_for_optional),
        fields=fields,
        omit_none=cls_none_format is OmitNone,
        is_frozen=_is_frozen_dataclass(t, fields[0]) if fields else False,
        doc=_get_dataclass_doc(t),
        custom_encoder=custom_encoder,
    )


def _get_entity_fields(t: Any) -> Sequence[_Field[Any]]:
    if dataclasses.is_dataclass(t):
        return [
            _Field(
                name=f.name,
                type=f.type,
                default=(DefaultValue.some(f.default) if f.default is not dataclasses.MISSING else NOT_SET),
                default_factory=(
                    DefaultValue.some(f.default_factory) if f.default_factory is not dataclasses.MISSING else NOT_SET
                ),
            )
            for f in dataclasses.fields(t)
        ]
    if is_typeddict(t):
        return [
            _Field(
                name=field_name,
                type=field_type,
                default=NOT_SET if _is_required_in_typeddict(t, field_name) else DefaultValue.some(None),
                default_factory=NOT_SET,
            )
            for field_name, field_type in t.__annotations__.items()
        ]
    if _is_attrs(t):
        assert attr
        return [
            _Field(
                name=f.name,
                type=f.type,
                default=(
                    DefaultValue.some(f.default)
                    if (
                        f.default is not attr.NOTHING and not isinstance(f.default, attr.Factory)  # type: ignore[arg-type]
                    )
                    else NOT_SET
                ),
                default_factory=(
                    DefaultValue.some(f.default.factory)  # pyright: ignore
                    if isinstance(f.default, attr.Factory)  # type: ignore[arg-type]
                    else NOT_SET
                ),
            )
            for f in attr.fields(t)
        ]

    raise RuntimeError(f"Unsupported type '{t}'")


@overload
def _find_metadata(annotations: Iterable[Any], type_: type[_T], default: _T) -> _T: ...


@overload
def _find_metadata(annotations: Iterable[Any], type_: type[_T], default: None = None) -> Optional[_T]: ...


def _find_metadata(annotations: Iterable[Any], type_: type[_T], default: Optional[_T] = None) -> Optional[_T]:
    return next((ann for ann in annotations if isinstance(ann, type_)), default)


def _wrap_annotated(annotations: Iterable[Any]) -> Callable[[_T], _T]:
    def inner(type_: _T) -> _T:
        for ann in annotations:
            type_ = Annotated[type_, ann]  # type: ignore
        return type_

    return inner


def _get_annotated_metadata(t: Any) -> tuple[Any, ...]:
    if get_origin(t) == Annotated:
        return getattr(t, '__metadata__', ())
    return ()


def _apply_format(f: Optional[FieldFormat], value: str) -> str:
    if not f or f.format is Format.no_format:
        return value
    if f.format is Format.camel_case:
        return to_camelcase(value)
    assert_never(f.format)


_NAME_CACHE = {}


@cache
def _generate_name(
    cls: Any,
    field_format: FieldFormat,
    none_format: NoneFormat,
    cls_none_as_default_for_optional: NoneAsDefaultForOptional,
) -> str:
    """
    Generate unique name for entity type.

    We need unique name to avoid name conflicts in generated json schema,
    when we have one entity with different Annotated metadata (like FieldFormat).
    """
    name = repr(cls).removeprefix("<class '").removesuffix("'>")
    if name not in _NAME_CACHE:
        _NAME_CACHE[name] = 0
        return name

    _NAME_CACHE[name] += 1
    return f'{name}{_NAME_CACHE[name]}'


def _get_globals(t: Any) -> dict[str, Any]:
    if t.__module__ in sys.modules:
        return sys.modules[t.__module__].__dict__.copy()
    return {}


def _evaluate_forwardref(t: type[_T], meta: Meta) -> type[_T]:
    if not isinstance(t, ForwardRef):
        return t

    return t._evaluate(meta.globals, {}, set())


def _get_discriminator_value(t: Any, name: str) -> str:
    fields = attr.fields(t) if attr and _is_attrs(t) else dataclasses.fields(t)
    for field in fields:
        if field.name == name:
            if _is_literal_type(field.type):
                args = get_args(field.type)
                if len(args) != 1:
                    raise RuntimeError(
                        f'Type {t} has invalid discriminator field "{name}". '
                        f'Discriminator supports only Literal[...] with one argument.'
                    )
                arg = args[0]

                if isinstance(arg, Enum):
                    arg = arg.value

                if isinstance(arg, str):
                    return arg

            raise RuntimeError(
                f'Type {t} has invalid discriminator field "{name}" with type "{field.type!r}". '
                f'Discriminator supports Literal[<str>], Literal[Enum] with str values.'
            )
    raise RuntimeError(f'Type {t} does not have discriminator field "{name}"')


def _is_str_literal(t: Any) -> bool:
    if _is_literal_type(t):
        args = get_args(t)
        if args and all(isinstance(arg, str) for arg in args):
            return True
    return False


def _is_literal_type(t: Any) -> bool:
    return t is Literal or get_origin(t) is Literal


def _is_attrs(t: Any) -> bool:
    return attr is not None and attr.has(t)


def _is_required_in_typeddict(t: Any, key: str) -> bool:
    if is_typeddict(t):
        if t.__total__:
            return key not in t.__optional_keys__
        return key in t.__required_keys__
    raise RuntimeError(f'Expected TypedDict, got "{t!r}"')


def _get_dataclass_doc(cls: Any) -> Optional[str]:
    """Dataclass has automatically generated docstring, which is not very useful."""
    doc: str = cls.__doc__
    if doc is None or doc.startswith(f'{cls.__name__}('):
        return None
    return doc


def _is_frozen_dataclass(cls: Any, field: EntityField) -> bool:
    try:
        obj = object.__new__(cls)
        setattr(obj, field.name, None)
    except FrozenInstanceErrors:
        return True
    else:
        return False


def _is_new_type(t: Any) -> bool:
    return hasattr(t, '__supertype__')



from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from typing import TypeVar, Union

from ._impl import CustomEncoder


@dataclass(frozen=True)
class Min:
    value: Union[int, float]


@dataclass(frozen=True)
class Max:
    value: Union[int, float]


@dataclass(frozen=True)
class MinLength:
    value: int


@dataclass(frozen=True)
class MaxLength:
    value: int


@dataclass(frozen=True)
class Discriminator:
    name: str


@dataclass(frozen=True)
class Alias:
    value: str


class Format(Enum):
    no_format = 'no_format'
    camel_case = 'camel_case'


@dataclass(frozen=True)
class FieldFormat:
    format: Format


CamelCase: FieldFormat = FieldFormat(Format.camel_case)
NoFormat: FieldFormat = FieldFormat(Format.no_format)


@dataclass(frozen=True)
class NoneFormat:
    omit: bool


KeepNone: NoneFormat = NoneFormat(False)
OmitNone: NoneFormat = NoneFormat(True)


@dataclass(frozen=True)
class NoneAsDefaultForOptional:
    use: bool


ForceDefaultForOptional: NoneAsDefaultForOptional = NoneAsDefaultForOptional(True)
KeepDefaultForOptional: NoneAsDefaultForOptional = NoneAsDefaultForOptional(False)


_I = TypeVar('_I')
_O = TypeVar('_O')


def serialize_with(func: Callable[[_I], _O]) -> CustomEncoder[_I, _O]:
    return CustomEncoder[_I, _O](serialize=func)


def deserialize_with(func: Callable[[_O], _I]) -> CustomEncoder[_I, _O]:
    return CustomEncoder[_I, _O](deserialize=func)
